LABELS_CONCAT = ":"
LABEL = "label"
EDGE_TYPE = "type"
DGL_ID = "dgl_id"
DGL_DEFAULT_NODE_LABEL = "_N"
DGL_DEFAULT_EDGE_TYPE = "_E"
PYG_ID = "pyg_id"
EDGE_INDEX = "edge_index"
NUM_NODES = "num_nodes"
DEFAULT_NODE_LABEL = "NODE"
DEFAULT_EDGE_TYPE = "RELATIONSHIP"
